#!/usr/bin/env python
# coding: utf8

"""
    Visual keyboard for pygame. Aims to be easy to use as highly
    customizable as well.
"""

from .vkeyboard import VKeyboardRenderer, VKeyboardLayout, VKeyboard

__version__ = '2.0.2'
